import numpy as np
from warnings import warn


def compute_forming_limit_curve(all_model_responses, strain_rate_ratio_threshold,
                                all_displacement_BCs, all_groove_angles):
    """
    Parameters
    ----------
    all_model_responses : list of dict
        Each list item must be a dict with keys:
            time
            LE_mid_mises
            LE_corner_mises
            LE_corner_22
            LE_corner_11
    strain_rate_ratio_threshold : number
    all_displacement_BCs : list of list
    all_groove_angles : list of float

    Returns
    -------
    forming_limit_curve : dict
        Dict with the following keys:
            strain_paths : list of list of ndarray
                Each list element is a sub-list corresponding to the strain paths of a 
                given groove angle. Each sub-list item is an array with two rows (minor
                and major strain respectively), representing the strain path up to
                necking. 
            forming_limits : ndarray
                Each column is the most conservative forming limit (minor and major
                strain) for a given strain path.

    """

    forming_limit_curve = {
        'displacement_BCs': all_displacement_BCs,
        'groove_angles': all_groove_angles,
    }

    strain_paths = []

    for resp, groove_angle, disp_BCs in zip(
        all_model_responses,
        all_groove_angles,
        all_displacement_BCs
    ):

        first_derivative = np.diff(resp['LE_11'], axis=0)

        # For all increments, identify elements with min/max strain rate:
        elem_idx_min = np.argmin(first_derivative, axis=1)
        elem_idx_max = np.argmax(first_derivative, axis=1)

        # For all increments, find min/max strain rate over all elements
        min_derivative = first_derivative[:, elem_idx_min]
        max_derivative = first_derivative[:, elem_idx_max]

        # For all increments, maximal strain rate ratio
        min_derivative[np.isclose(min_derivative, 0)] = np.nan
        strain_rate_ratio = np.abs(max_derivative / min_derivative)

        strain_path_i = {
            'groove_angle_deg': groove_angle,
            'displacement_BCs': disp_BCs,
            'strain_rate_ratio': strain_rate_ratio,
            'strain': None,
        }
        strain_paths.append(strain_path_i)

        try:
            first_threshold_idx = np.where(
                strain_rate_ratio >= strain_rate_ratio_threshold
            )[0][0]
        except IndexError:
            warn(f'Simulation did not reach the target threshold value '
                 f'({strain_rate_ratio_threshold}) for groove angle '
                 f'{groove_angle} (degrees) and displacment BCs: {disp_BCs}.')
            continue

        idx_min = elem_idx_min[first_threshold_idx]

        first_threshold_idx = first_threshold_idx + 1

        minor_strain = resp['LE_22'][:first_threshold_idx + 1, idx_min]
        major_strain = resp['LE_11'][:first_threshold_idx + 1, idx_min]
        strain_path = np.vstack((minor_strain, major_strain))
        strain_paths[-1].update({'strain': strain_path})

    forming_limit_curve.update({'strain_paths': strain_paths})

    # Get the overal forming limit for each strain path, taking the smallest major strain
    # over all groove angles for a given strain path:
    forming_limits = []
    seen_disp_BCs = []
    selected_angles = []
    for disp_BC in all_displacement_BCs:

        if disp_BC not in seen_disp_BCs:
            seen_disp_BCs.append(disp_BC)
        else:
            continue

        strain_paths_sub = [i for i in strain_paths if i['displacement_BCs'] == disp_BC]

        if all([i['strain'] is None for i in strain_paths_sub]):
            warn(f'No simulations (at any groove angle) reached the target threshold '
                 f'value for disp_BC: {disp_BC}.')
            forming_limits.append([np.nan, np.nan])
            selected_angles.append(np.nan)
            continue

        all_max_maj = []
        # Looping over groove angles:
        for strain_path_i in strain_paths_sub:
            if strain_path_i['strain'] is None:
                all_max_maj.append(np.inf)
            else:
                all_max_maj.append(np.max(strain_path_i['strain'][1]))

        safest_angle_idx = np.argmin(all_max_maj)
        forming_limits.append(strain_paths_sub[safest_angle_idx]['strain'][:, -1])
        selected_angles.append(strain_paths_sub[safest_angle_idx]['groove_angle_deg'])

    forming_limits = np.array(forming_limits)
    selected_angles = np.array(selected_angles)
    srt_idx = np.argsort(forming_limits[:, 0])  # Sort by minor strain
    final_forming_limits = forming_limits[srt_idx].T  # Transpose to column vectors
    selected_angles = selected_angles[srt_idx]

    forming_limit_curve.update({
        'forming_limits': final_forming_limits,
        'forming_limit_groove_angles_deg': selected_angles,
    })

    return forming_limit_curve



import numpy as np


def prepare_Hill1948_parameters(parameters):
    """Prepare the parameter list for the Hill1948 yield function."""

    R33 = np.sqrt(1 / (parameters['F'] + parameters['G']))
    R22 = np.sqrt(1 / (parameters['F'] + parameters['H']))
    R11 = np.sqrt(1 / (parameters['G'] + parameters['H']))

    R23 = np.sqrt(3 / (2 * parameters['L']))
    R13 = np.sqrt(3 / (2 * parameters['M']))
    R12 = np.sqrt(3 / (2 * parameters['N']))

    return [R11, R22, R33, R12, R13, R23]


def prepare_Barlat_Yld91_parameters(parameters):
    """Prepare the parameter list for the Barlat_Yld91 yield function."""
    return [
        parameters['a'],
        parameters['b'],
        parameters['c'],
        parameters['f'],
        parameters['g'],
        parameters['h'],
    ]


def prepare_Barlat_Yld2004_18p_parameters(parameters):
    """Prepare the parameter list for the Barlat_Yld2004_18p yield function."""
    return [
        parameters['c_p_12'],
        parameters['c_p_13'],
        parameters['c_p_21'],
        parameters['c_p_23'],
        parameters['c_p_31'],
        parameters['c_p_32'],
        parameters['c_p_66'] * 0.5,
        parameters['c_p_55'] * 0.5,
        parameters['c_p_44'] * 0.5,
        parameters['c_dp_12'],
        parameters['c_dp_13'],
        parameters['c_dp_21'],
        parameters['c_dp_23'],
        parameters['c_dp_31'],
        parameters['c_dp_32'],
        parameters['c_dp_66'] * 0.5,
        parameters['c_dp_55'] * 0.5,
        parameters['c_dp_44'] * 0.5,
    ]


YIELD_FUNC_LOOKUP = {
    'Barlat_Yld91': prepare_Barlat_Yld91_parameters,
    'Barlat_Yld2004_18p': prepare_Barlat_Yld2004_18p_parameters,
    'Hill1948': prepare_Hill1948_parameters,
}



from abaqus_parse import write_MK_mesh


def generate_MK_mesh(name_inp, FE_input_data):

    ######    Sample data   ######
    sample_size = FE_input_data['sample_size']
    # Groove geometry
    # thickness of the groove
    Inhomogeneity_factor = FE_input_data['inhomogeneity_factor']
    L_groove = FE_input_data['L_groove']
    L_slope = FE_input_data['L_slope']
    # Material angle
    Material_angle = FE_input_data['material_angle']  # (°)
    # Groove angle
    Groove_angle = 90 - FE_input_data['groove_angle']  # (°)
    # Material properties
    E = FE_input_data['elastic_modulus']
    mu = FE_input_data['poisson_ratio']
    rho = FE_input_data['density']

    law = FE_input_data['law']

    Plastic = FE_input_data['plastic']

    ######    FE data   ######
    mesh_size = FE_input_data['mesh_size']
    Element_type = FE_input_data['elem_type']
    Nb_el_thickness = FE_input_data['Nb_el_thickness']
    # Strain rate (/s)
    Eps_rate = FE_input_data['strain_rate']

    # Step time (put 0 => default Abaqus value)
    time_step = FE_input_data['total_time']  # Time period of the step (s)
    dt_i = FE_input_data['time_step'][0]  # Suggested initial time increment (s)
    dt_min = FE_input_data['time_step'][1]  # Minimum time increment allowed (s)
    dt_max = FE_input_data['time_step'][2]  # Maximum time increment allowed (s)
    # Bulk viscosity
    b1 = FE_input_data['bulk_parameters'][0]  # Linear bulk viscosity parameter
    b2 = FE_input_data['bulk_parameters'][1]  # Quadratic bulk viscosity parameter
    # Boundary conditions
    U_left = FE_input_data['displacment_BC'][0]  # (mm) along x
    U_right = FE_input_data['displacment_BC'][1]  # (mm) along x
    U_up = FE_input_data['displacment_BC'][2]  # (mm) along y
    U_bottom = FE_input_data['displacment_BC'][3]  # (mm) along y
    Max_plastic_strain = FE_input_data['max_plastic_strain']  # Eq. plastic strain PEEQ
    num_interval = FE_input_data.get('num_interval')

    # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    #                             Writing
    # #############################################################################
    write_MK_mesh.heading(name_inp, Inhomogeneity_factor,
                          Material_angle, Groove_angle, Eps_rate, law)

    write_MK_mesh.part(name_inp, sample_size, mesh_size, Inhomogeneity_factor,
                       L_groove, L_slope, Element_type, Material_angle, Groove_angle, Nb_el_thickness)

    write_MK_mesh.amplitude(name_inp, Eps_rate)

    write_MK_mesh.material(name_inp, E, mu, rho, Plastic, law)

    write_MK_mesh.step(name_inp, time_step, dt_i, dt_min, dt_max, b1, b2, U_left,
                       U_right, U_up, U_bottom, Max_plastic_strain, num_interval)



import numpy as np


def save_model_response(fname):

    data = np.loadtxt(fname, comments='%')

    nb_el = int(np.where(data[:, 0] > 0.0)[0][0])
    nb_frame = int(np.shape(data)[0]/nb_el)

    time = data[0:-1:nb_el, 0]
    LE_mises = np.reshape(data[:, 2], [nb_frame, nb_el])
    LE_11 = np.reshape(data[:, 3], [nb_frame, nb_el])
    LE_22 = np.reshape(data[:, 4], [nb_frame, nb_el])

    model_response = {
        'time': time,
        'LE_mises': LE_mises,
        'LE_11': LE_11,
        'LE_22': LE_22,
    }

    return model_response



from plotly import graph_objs as go
def plot_nodes(plot_x, plot_y, plot_z=None, indices=None, xrange=None, yrange=None,
               figsize=[400, 400], dticks=None, mode='markers+text',
              markersize=5, showlegend=False, show_labels=True, threed=False):
    """
    Parameters
    ----------
    show_labels : bool, optional
        If True, show node labels.
        
    """
    data = []
    if threed:
        if type(plot_x) == list:
            i = 0
            for plx, ply, plz in zip(plot_x, plot_y, plot_z):
                if indices == None:
                    text = None
                else:
                    text = indices[i]
                trace = {
                    'type': 'scatter3d',
                    'x': plx,
                    'y': ply,
                    'z': plz,
                     'mode': mode,
                    'marker': {'size': markersize},                
                }
                if show_labels:
                    trace.update({'text': text, 'type': 'scatter3d'})

                data.append(trace)

                i += 1
        else:    
            if indices is not None:
                indices = indices
                print('here')
                print('indices shape: ', indices.shape)
                print('plot_x shape: ', plot_x.shape)
                data.append({
                     'type': 'scatter3d',
                        'x': plot_x,
                        'y': plot_y,
                        'z': plot_z,
                         'mode': mode,
                       'marker': {'size': markersize},
                        'text': indices
                    })
            else:
                data.append({
                     'type': 'scatter3d',
                        'x': plot_x,
                        'y': plot_y,
                        'z': plot_z,
                        'text': None,
                        'mode': 'markers',
                        'marker': {'size': markersize},
                    })
    else:
        if type(plot_x) == list:
            i = 0
            for plx, ply in zip(plot_x,plot_y):
    #             print(i, plot_x[i])
                if indices == None:
                    text = None
                else:
    #                 print(i)
                    text = indices[i]
    #                 print(indices[i])
                trace = {
                    'type': 'scattergl',
                    'x': plx,
                    'y': ply,
                     'mode': mode,
                    'marker': {'size': markersize},                
                }
                if show_labels:
                    trace.update({'text': text, 'type': 'scatter'})

                data.append(trace)

                i += 1
        else:    
            if indices is not None:
                indices = indices
                data.append({
                        'x': plot_x,
                        'y': plot_y,
                         'mode': mode,
                        'marker': {'size': markersize},
                        'text': indices
                    })
            else:
                data.append({
                        'x': plot_x,
                        'y': plot_y,
                        'text': None,
                         'mode': 'markers',
                        'marker': {'size': markersize},
                    })

    layout = {
        'width': figsize[0],
        'height': figsize[1],
        'xaxis': {
    #         'range': [-0.1, 1.1],
            'zeroline': False,
            'showline': True,
            'mirror': 'all',
    #         'title': r'$m$',
            'titlefont':{'size': 13},
        },
        'yaxis': {
            'zeroline': False,
            'showline': True,
            'mirror': 'all',
    #         'title': '$\sigma_\mathrm{w}/\sigma_\mathrm{f} $ ',
            'titlefont':{'size': 13},
            'scaleanchor': 'x',
            'scaleratio': 1,
        },
        'annotations':[
        {
            'text':  r'$\LARGE{x}$',
            'x': 0.5,
            'y': -0.1,
            'xref': 'paper',
            'yref': 'paper',
            'showarrow': False,
            'font': {'size': 11},
        },
            {
            'text':  r'$\LARGE{y}$',
            'x': -0.1,
            'y': 0.5,
            'xref': 'paper',
            'yref': 'paper',
            'showarrow': False,
            'font': {'size': 11},
            'textangle': -90
        },
        ],
        'legend':{
            'x': 0.6,
            'y': 0.7,
        },
         'margin': {
                    'l': 80,
                    'r': 20,
                    'b': 80,
                    't': 40,
    #                 'pad': 1
         },

    }
    if xrange:
        layout['xaxis']['range'] = xrange
    if yrange:
        layout['yaxis']['range'] = yrange
    if dticks:
        layout['xaxis']['dtick'] = dticks[0]
        layout['yaxis']['dtick'] = dticks[1]
    layout['showlegend'] = showlegend
    f = go.FigureWidget(data)
    return f
